/**************************************************************************************************
  GPS_Click.c
  GPS application.

  The GPS Click board is located in slot 2 of the MikroMedia+ for STM32 shield for this code.

  GPS Name         mikroBUS      Slot 2 Name      HDR1    HDR2      MM Schematic      STM32 Pin
---------------------------------------------------------------------------------------------------
  Unused              AN         PB0/AN2           4                HDR-AN1           PB0   46
  RESET_N (output)    RST        PE4               9                HDR-GPIO1         PE4   3
  Unused              CS         PC13              10               HDR-GPIO2         PC13  7
  Unused              SCK        PA5/SPI-SCK       18               HDR-SPI1_SCK      PA5   41
  Unused              MISO       PA6/SPI-MISO      19               HDR-SPI1_MISO     PA6   42
  Unused              MOSI       PB5/SPI-MOSI      20               HDR-SPI1_MOSI     PB5   135
  +3.3V (power)       +3.3V      VCC-3.3V          25      28       VCC-3.3V
  GND (ground)        GND                          2,26

  Unused              PWM        PD13/PWM2                 34       HDR-PWM1          PD13  82
  TIMEPULSE (input)   INT        PE1/INT2                  40       HDR-INT1          PE1   142
  TxD1 (alt func)     TX         PA3/UART1-RX              47       HDR-UART1_RX      PA3   37
  RxD1 (alt func)     RX         PD5/UART1-TX              48       HDR-UART1_TX      PD5   119
  SCL2                SCL        PF1/IC2-SCL               43       HDR-I2C2_SCL      PF1   11
  SDA2                SDA        PF0/IC2-SDA               44       HDR-I2C2_SDA      PF0   10
  Unused              +5V        VCC-5V            1                VCC-5V
  GND (ground)        GND                                  51



  NOTE: Initially, the GPS module takes about 30 seconds to acquire satellites. It begins in
        verbose mode by sending all of the packets. You can turn off verbose mode by sending
        command 'd' on UART6. Command 'e' turns verbose mode back on.

**************************************************************************************************/

// Include Files
#include "GPS_Click.h"
#include "Buffers.h"
#include "UART6.h"
#include "UART2.h"
#include "Timers.h"
#include "__Time.h"
#include "RTC.h"
#include "Common.h"
#include "GPS_resources.h"
#include "GPS_objects.h"

// GPS click board connections
sbit GPS_RST at GPIOE_ODR.B4;
sbit GPS_TIMEPULSE at GPIOE_IDR.B1;
sbit GPS_SCL at GPIOF_IDR.B1;
sbit GPS_SDA at GPIOF_IDR.B0;



// Variables
char GPS_State = 0;
char DecodeBuffer[256];

char ID[8];
char Time[16];
char Day[4];
char Month[4];
char Year[8];
char Hour[4];
char Minute[4];
char Second[4];

GPS_GPGGA GGA;   // Buffers for GGA message

TimeStruct MyTS;
long Epoch;
char WeekDay[16];
char GPS_Count = 0;
char UpdateFlag = 0;

unsigned short int GPS_TimerHandle;



// From mikroE example
char *string;
int latitude, longitude, oldlatitude, oldlongitude;
char lat_str[8];
char long_str[9];



// External variables
extern char U2_MsgFlg;
extern char U2_RcvBuf[];
extern TTime MyTime;






/**************************************************************************************************
  External interrupts
**************************************************************************************************/
/*
void Test0_Interrupt() iv IVT_INT_EXTI0 ics ICS_AUTO {     // Unused
     PR0_bit = 1;  // Clear interrupt
     UART6_PutQueue("Int0\r\n");
}
*/

void GPS_Interrupt() iv IVT_INT_EXTI1 ics ICS_AUTO {       //  Interrupt from GPS
     PR1_bit = 1;  // Clear interrupt
     //UART6_PutQueue("Int1\r\n");  // Test
     GPS_Count += 1;
     if (GPS_Count > 9) {   // Update every 10 seconds
         GPS_Count = 0;
         UpdateFlag = 1;
     }
}

/*
void Test2_Interrupt() iv IVT_INT_EXTI2 ics ICS_AUTO {     // Unused
     PR2_bit = 1;  // Clear interrupt
     UART6_PutQueue("Int2\r\n");
}

void Test3_Interrupt() iv IVT_INT_EXTI3 ics ICS_AUTO {     // Unused
     PR3_bit = 1;  // Clear interrupt
     UART6_PutQueue("Int3\r\n");
}

void Test4_Interrupt() iv IVT_INT_EXTI4 ics ICS_AUTO {     // Unused
     PR4_bit = 1;  // Clear interrupt
     UART6_PutQueue("Int4\r\n");
}
*/

/**************************************************************************************************
  GPS Initialize
**************************************************************************************************/
void GPS_Initialize(void) {

     // Not required when using mikroE libraries
     //GPIO_Clk_Enable(&GPIOA_BASE);                          // Enable clock on port A
     //GPIO_Clk_Enable(&GPIOB_BASE);                          // Enable clock on port B
     //GPIO_Clk_Enable(&GPIOC_BASE);                          // Enable clock on port C
     //GPIO_Clk_Enable(&GPIOD_BASE);                          // Enable clock on port D
     //GPIO_Clk_Enable(&GPIOE_BASE);                          // Enable clock on port E
     //GPIO_Clk_Enable(&GPIOF_BASE);                          // Enable clock on port F


     GPIO_Digital_Output(&GPIOE_BASE,_GPIO_PINMASK_4);      // GPS_RST is an output
     GPIO_Digital_Input(&GPIOE_BASE,_GPIO_PINMASK_1);       // GPS_TIMEPULSE is an input
     GPIO_Digital_Input(&GPIOF_BASE,_GPIO_PINMASK_1);       // GPS_SCL is an input
     GPIO_Digital_Input(&GPIOF_BASE,_GPIO_PINMASK_0);       // GPS_SDA is an input


     // Reset the GPS click module
     GPS_RST = 0;     // Activate the reset signal
     Delay_ms(10);
     GPS_RST = 1;     // Deactivate the reset signal

     // Set EXTI_IMR register
     MR1_bit = 1;     // Unmask interrupt request from line 1

     // Set EXTI_RTSR register
     TR1_bit = 1;     // Enable rising trigger
     
     // Set SYSCFGEN bit in RCC_APB2ENR register
     SYSCFGEN_bit = 1;  // Enable system configuration controller clock
     
     // Set SYSCFG_EXTICR1 bit 6
     SYSCFG_EXTICR1 = 0x00000040;    // Connect PE1 to EXTI1

     NVIC_IntEnable(IVT_INT_EXTI1);  // Enable EXTI1 interrupt

     
}

/**************************************************************************************************
  Copied from mikroE example
**************************************************************************************************/
void Display_Cursor(signed int lat, signed int lon, char *lt_Str, char *lng_Str, unsigned int color) {
     unsigned long latitude_y, longitude_x;

     // Latitude and Longitude scaling for 320x240 display:
     // Latitude: Input range is -90 to 90 degrees
     // Longitude: Input range is -180 to 180 degrees
     latitude_y = ((160*(90 - lat))/180) + 40;
     longitude_x = (320*(lon + 180));
     longitude_x = longitude_x/360 - 10;

     // Cursor drawing
     TFT_Set_Pen(CL_RED, 1);
     TFT_H_Line(longitude_x - 2, longitude_x + 2, latitude_y );
     TFT_V_Line( latitude_y - 2, latitude_y + 2, longitude_x );

     TFT_Set_Font(&TFT_defaultFont, color, FO_HORIZONTAL);
     TFT_Write_Text(lt_Str, 160, 200);
     TFT_Write_Text(lng_Str, 160, 215);
}

/**************************************************************************************************
  Decode response subroutine
**************************************************************************************************/
void DecodeResponse(void) {
     char *BufNum;
     
     if ((DecodeBuffer[1] == 'G') && (DecodeBuffer[2] == 'P')) { // All messages start with $GP
         if ((DecodeBuffer[3] == 'Z') && (DecodeBuffer[4] == 'D') && (DecodeBuffer[5] == 'A')) {
             DEBUGOUT("Received ZDA\r\n");  // Diagnostic
             GetToken(DecodeBuffer,ID);
             GetToken(DecodeBuffer,Time);
             GetToken(DecodeBuffer,Day);
             GetToken(DecodeBuffer,Month);
             GetToken(DecodeBuffer,Year);
             Hour[0] = Time[0];
             Hour[1] = Time[1];
             Hour[2] = 0;
             Minute[0] = Time[2];
             Minute[1] = Time[3];
             Minute[2] = 0;
             Second[0] = Time[4];
             Second[1] = Time[5];
             Second[2] = 0;
/*
             // Diagnostic
             BufNum = GetBuffer();
             sprintf(BufNum,"%s/%s/%s %s:%s:%s\r\n",Year,Month,Day,Hour,Minute,Second);
             UART6_PutQueue(BufNum);
*/
             MyTS.yy = atoi(Year);
             MyTS.mo = atoi(Month);
             MyTS.md = atoi(Day);
             MyTS.hh = atoi(Hour);
             MyTS.mn = atoi(Minute);
             MyTS.ss = atoi(Second);
             
             Epoch = Time_dateToEpoch(&MyTS);
             Epoch -= 14400;                     // Adjust for Greenwich Mean Time (GMT)
             Time_epochToDate(Epoch, &MyTS);
             
             switch (MyTS.wd) {
                 case 0:
                     strcpy(WeekDay,"Monday");
                     break;
                 case 1:
                     strcpy(WeekDay,"Tuesday");
                     break;
                 case 2:
                     strcpy(WeekDay,"Wednesday");
                     break;
                 case 3:
                     strcpy(WeekDay,"Thursday");
                     break;
                 case 4:
                     strcpy(WeekDay,"Friday");
                     break;
                 case 5:
                     strcpy(WeekDay,"Saturday");
                     break;
                 case 6:
                     strcpy(WeekDay,"Sunday");
                     break;
             }
#if DEBUG == 1
             // Display date/time in the proper notation
             BufNum = GetBuffer();
             sprintf(BufNum,"GPS: %04d/%02d/%02d %02d:%02d:%02d %s\r\n",MyTS.yy,MyTS.mo,MyTS.md,MyTS.hh,MyTS.mn,MyTS.ss,WeekDay);
             UART6_PutQueue(BufNum);
#endif
         
             UpdateLabel1(BufNum,CL_BLACK);  // Update TFT label
         
         
         }
         
         if ((DecodeBuffer[3] == 'G') && (DecodeBuffer[4] == 'G') && (DecodeBuffer[5] == 'A')) {
             DEBUGOUT("Received GGA\r\n");  // Diagnostic
             GetToken(DecodeBuffer,GGA.ID);
             GetToken(DecodeBuffer,GGA.Time);
             GetToken(DecodeBuffer,GGA.Latitude);
             GetToken(DecodeBuffer,GGA.NS);
             GetToken(DecodeBuffer,GGA.Longitude);
             GetToken(DecodeBuffer,GGA.EW);
             GetToken(DecodeBuffer,GGA.Fix);
             GetToken(DecodeBuffer,GGA.NumSat);
             GetToken(DecodeBuffer,GGA.HDOP);
             GetToken(DecodeBuffer,GGA.Altitude);
             GetToken(DecodeBuffer,GGA.FF1);
             GetToken(DecodeBuffer,GGA.Height);
             GetToken(DecodeBuffer,GGA.FF2);
             GetToken(DecodeBuffer,GGA.Age);
             GetToken(DecodeBuffer,GGA.DGPS_ID);
             
             BufNum = GetBuffer();
             sprintf(BufNum,"%s;%s;%s;%s;%s;%s;%s;%s\r\n",GGA.ID,GGA.Time,GGA.Latitude,GGA.NS,GGA.Longitude,GGA.EW,GGA.Fix,GGA.NumSat);
             UART6_PutQueue(BufNum);
         }
     
         if ((DecodeBuffer[3] == 'G') && (DecodeBuffer[4] == 'L') && (DecodeBuffer[5] == 'L')) {
             DEBUGOUT("Received GLL\r\n");  // Diagnostic

             // Copied from mikroE example
             string = strstr(U2_RcvBuf,"$GPGLL");
             
             latitude = (string[7]-48)*10 + (string[8]-48);
             longitude = (string[19]-48)*100 + (string[20]-48)*10 + (string[21]-48);
             if (string[17] == 'S') {            // if the latitude is in the South direction it has minus sign
                 latitude = 0 - latitude;
             }
            
             if (string[30] == 'W') {            // if the longitude is in the West direction it has minus sign
                 longitude = 0 - longitude;
             }

            if ((oldlongitude != longitude) | (oldlatitude != latitude)){
                Display_Cursor(latitude, longitude, &lat_str, &long_str, CL_WHITE); // Display the cursor on the world map

                lat_str[0] = string[7];
                lat_str[1] = string[8];
                lat_str[2] = '-';
                lat_str[3] = string[9];
                lat_str[4] = string[10];
                lat_str[5] = ' ';
                lat_str[6] = string[17];
                lat_str[7] = 0;

                long_str[0] = string[19];
                long_str[1] = string[20];
                long_str[2] = string[21];
                long_str[3] = '-';
                long_str[4] = string[22];
                long_str[5] = string[23];
                long_str[6] = ' ';
                long_str[7] = string[30];
                long_str[8] = 0;
             
                Display_Cursor(latitude, longitude, &lat_str, &long_str, CL_RED);   // Display the cursor on the world map
             }
         }
     }
}

/**************************************************************************************************
  This subroutine checks for a message received by UART2 by testing U2_MsgFlg. If it finds a
  message, it clears the flag and outputs the message on UART6. Returns -1 if there was no
  response, 0 if the received message was too long (prevents buffer overrun), 1 if a response
  was received properly.
**************************************************************************************************/
short WaitForResponse(void) {
     char *BufNum;

     if (U2_MsgFlg) {
         U2_MsgFlg = 0;

         if (strlen(U2_RcvBuf) > 255) {
             UART6_PutQueue("Message received from GPS is too long\r\n");
             return 0;  // Error
         } else {
             strcpy(DecodeBuffer,U2_RcvBuf);
#if DEBUG == 1
             BufNum = GetBuffer();
             strcpy(BufNum,U2_RcvBuf);
             UART6_PutQueue(BufNum);
#endif
             return 1;  // Received response
         }
     } else {
         return -1;     // No response
     }
}

/**************************************************************************************************
  GPS state machine
**************************************************************************************************/
void GPS_TaskMgr(void) {
     char *BufNum;
     short ResponseCode;

     switch (GPS_State) {

     /******* Hold here and scan for usolicited responses from GPS *******************************/
         case 0:
             ResponseCode = WaitForResponse();  // Output any unsolicited responses
     
     /****** Enable to update date/time every 10 seconds ******************************************
             if (UpdateFlag) {                  // Start the date/time update sequence
                 UpdateFlag = 0;
                 GPS_State = 1;
             }
     *********************************************************************************************/
             break;

     
     
     /******* This sequence will run when command 'u' is received on UART6 ***********************/
         case 1:
             DrawScreen(&Screen1);
             
             SendToGPS("$PMTK448,7*");          // Request ZDA message

             GPS_TimerHandle = get_timer();     // Get a free timer
             set_timer(GPS_TimerHandle,1000);   // Set timer for 1 second
             GPS_State = 2;
             break;
         case 2:
             ResponseCode = WaitForResponse();
             if (ResponseCode > 0) {
                 DecodeResponse();              // Decode the response
                 GPS_State = 3;
             }
             if (ResponseCode == 0) {
                 // Message is issued by WaitForResponse function
                 free_timer(GPS_TimerHandle);   // Return timer to pool
                 GPS_State = 0;
             }
             if (check_timer(GPS_TimerHandle) == 0) {
                 UART6_PutQueue("Timeout\r\n");
                 free_timer(GPS_TimerHandle);   // Return timer to pool
                 GPS_State = 0;
             }
             break;
         case 3:
             MyTime.year = MyTS.yy;             // Trasfer Time and date information
             MyTime.month = MyTS.mo;
             MyTime.day = MyTS.md;
             MyTime.hours = MyTS.hh;
             MyTime.minutes = MyTS.mn;
             MyTime.seconds = MyTS.ss;
             MyTime.ampm = 0;
             RTC_Write(&MyTime);                // Write time and date into RTC
             GPS_State = 4;
             break;
         case 4:
             /*
             RTC_Read(&MyTime);                 // Read time and date from RTC
             BufNum = GetBuffer();
             sprintf(BufNum,"RTC: 20%02d/%02d/%02d %02d:%02d:%02d\r\n",MyTime.Year,MyTime.month,MyTime.day,MyTime.hours,MyTime.minutes,MyTime.seconds);
             UART6_PutQueue(BufNum);            // Display it
             */
             free_timer(GPS_TimerHandle);       // Return timer to pool
             GPS_State = 0;
             break;
     
     /******* This sequence will run when command 'm' is received on UART6 ***********************/
     /******* Partially copied from mikroE example ***********************************************/
         case 5:
             TFT_Fill_Screen(CL_WHITE);
             TFT_Image(0, 40, WorldImage_bmp, 1);

             //TFT_Set_Font(&HandelGothic_BT21x22_Regular, CL_RED, FO_HORIZONTAL);
             TFT_Set_Font(&Tahoma21x25_Regular, CL_RED, FO_HORIZONTAL);
             TFT_Write_Text("GPS click Example", 40, 15);
             TFT_Set_Font(&TFT_defaultFont, CL_BLACK, FO_HORIZONTAL);
             TFT_Write_Text("Your latitude is:", 40, 200);
             TFT_Write_Text("Your longitude is:", 40, 215);
             
             GPS_State = 6;
             break;
         case 6:
            SendToGPS("$PMTK448,1*");          // Request GLL message

            GPS_TimerHandle = get_timer();     // Get a free timer
            set_timer(GPS_TimerHandle,1000);   // Set timer for 1 second
            GPS_State = 7;
            break;
         case 7:
            ResponseCode = WaitForResponse();
            if (ResponseCode > 0) {
                DecodeResponse();              // Decode the response
                GPS_State = 8;
            }
            if (ResponseCode == 0) {
                // Message is issued by WaitForResponse function
                free_timer(GPS_TimerHandle);   // Return timer to pool
                GPS_State = 0;
            }
            if (check_timer(GPS_TimerHandle) == 0) {
                UART6_PutQueue("Timeout\r\n");
                free_timer(GPS_TimerHandle);   // Return timer to pool
                GPS_State = 0;
            }
            break;
         case 8:
            free_timer(GPS_TimerHandle);       // Return timer to pool
            GPS_State = 0;
            break;
     
     /******* This sequence will run when command 'n' is received on UART6 ***********************/
         case 9:
             SendToGPS("$PMTK448,4*");          // Request GGA message

             GPS_TimerHandle = get_timer();     // Get a free timer
             set_timer(GPS_TimerHandle,1000);   // Set timer for 1 second
             GPS_State = 10;
             break;
         case 10:
             ResponseCode = WaitForResponse();
             if (ResponseCode > 0) {
                 DecodeResponse();              // Decode the response
                 GPS_State = 11;
             }
             if (ResponseCode == 0) {
                 // Message is issued by WaitForResponse function
                 free_timer(GPS_TimerHandle);   // Return timer to pool
                 GPS_State = 0;
             }
             if (check_timer(GPS_TimerHandle) == 0) {
                 UART6_PutQueue("Timeout\r\n");
                 free_timer(GPS_TimerHandle);   // Return timer to pool
                 GPS_State = 0;
             }
             break;
         case 11:
            free_timer(GPS_TimerHandle);       // Return timer to pool
            GPS_State = 0;
            break;

         
         
         // Place other useful sequences here

     }
}

/**************************************************************************************************
  Assuming the string s has the GPS message from $ to *, this function calculates the checksum
  and appends it to the message then sends it to the GPS click board using UART2.
**************************************************************************************************/
void SendToGPS(char *s) {
     char Checksum;
     char i;
     char *BufNum;
     char Buf1[8];

     // Calculate checksum
     Checksum = 0;
     for (i=1; i<strlen(s)-1; i++) {
          Checksum ^= s[i];
     }
     
     // Append checksum
     sprintf(Buf1,"%02X\r\n",Checksum);
     BufNum = GetBuffer();
     strcpy(BufNum,s);
     strcat(BufNum,Buf1);

     //UART6_PutQueue(BufNum);  // Test
     UART2_PutQueue(BufNum);
}

/**************************************************************************************************
  Given a string s that contains tokens delimited by a comma, this function strips the first
  token from the string and puts it in t. The return code is -1 if the string is empty, 0 if there
  are no more tokens following this one, 1 if there are more tokens to process. String s is
  destroyed in the process.
**************************************************************************************************/
short GetToken(char *s, char *t) {
     char *p;                       // Token pointer

     if (strlen(s) == 0) {          // If the string is empty
         return (-1);
     }
     p = strchr(s,',');             // Find location of comma
     if (p == NUL) {                // If didn't find a comma
         strcpy(t,s);               // Copy the final token
         strcpy(s,"");              // Nullify the input string
         return 0;
     }
     s[strlen(s)-strlen(p)] = '\0'; // Replace the comma with a null
     strcpy(t,s);                   // Copy the token
     p++;                           // Advance the pointer to the next character position
     strcpy(s,p);                   // Overlay the input string minus the first token
     return 1;
}


/**************************************************************************************************
  EOF
**************************************************************************************************/